package eu.jonahbauer.raytracing.scene.hittable2d;

import eu.jonahbauer.raytracing.math.AABB;
import eu.jonahbauer.raytracing.math.Range;
import eu.jonahbauer.raytracing.math.Ray;
import eu.jonahbauer.raytracing.math.Vec3;
import eu.jonahbauer.raytracing.render.material.Material;
import eu.jonahbauer.raytracing.scene.Target;
import eu.jonahbauer.raytracing.scene.util.PdfUtil;
import org.jetbrains.annotations.NotNull;

import java.util.random.RandomGenerator;

public final class Parallelogram extends Hittable2D implements Target {
    private final @NotNull AABB bbox;

    public Parallelogram(@NotNull Vec3 origin, @NotNull Vec3 u, @NotNull Vec3 v, @NotNull Material material) {
        super(origin, u, v, material);
        this.bbox = new AABB(origin, origin.plus(u).plus(v));
    }

    @Override
    protected boolean isInterior(double alpha, double beta) {
        return 0 <= alpha && alpha < 1 && 0 <= beta && beta < 1;
    }

    @Override
    public @NotNull AABB getBoundingBox() {
        return bbox;
    }

    @Override
    public double getProbabilityDensity(@NotNull Vec3 origin, @NotNull Vec3 direction) {
        if (Double.isNaN(hit0(new Ray(origin, direction), FORWARD))) return 0;

        var a = this.origin.minus(origin).unit();
        var b = this.origin.plus(u).minus(origin).unit();
        var c = this.origin.plus(v).minus(origin).unit();
        var d = this.origin.plus(u).plus(v).minus(origin).unit();
        var angle = PdfUtil.getSolidAngle(a, b, d) + PdfUtil.getSolidAngle(c, b, d);
        return 1 / angle;
    }

    @Override
    public @NotNull Vec3 getTargetingDirection(@NotNull Vec3 origin, @NotNull RandomGenerator random) {
        var alpha = random.nextDouble();
        var beta = random.nextDouble();
        return this.origin.plus(u.times(alpha)).plus(v.times(beta)).minus(origin);
    }
}
